from airiam.terraform.entity_terraformers.BaseEntityTransformer import BaseEntityTransformer
from airiam.terraform.entity_terraformers.IAMPolicyDocumentTransformer import IAMPolicyDocumentTransformer


class IAMPolicyTransformer(BaseEntityTransformer):
    def __init__(self, entity_json):
        super().__init__('aws_iam_policy', entity_json['PolicyName'], entity_json)
        self._policy_arn = entity_json['Arn']

    def _generate_hcl2_code(self, entity_json) -> str:
        document = next(version['Document'] for version in entity_json['PolicyVersionList'] if version['IsDefaultVersion'])
        policy = IAMPolicyDocumentTransformer(document, self._safe_name)
        tags = BaseEntityTransformer.transform_tags(entity_json)
        policy_code = f"""{policy.code()}

resource "aws_iam_policy" "{self._safe_name}" {{
  name        = "{entity_json['PolicyName']}"
  path        = "{entity_json['Path']}"
  policy      = {policy.identifier()}.json
  description = \"{entity_json['Description']}\"
}}

"""
        return policy_code

    def entities_to_import(self) -> list:
        return [{"identifier": self.identifier(), "entity": self._policy_arn}]



from airiam.terraform.entity_terraformers.BaseEntityTransformer import BaseEntityTransformer, Principal
from airiam.terraform.entity_terraformers.IAMInlinePolicyTransformer import IAMInlinePolicyTransformer
from airiam.terraform.entity_terraformers.IAMManagedPolicyAttachmentTransformer import IAMManagedPolicyAttachmentTransformer


class IAMUserTransformer(BaseEntityTransformer):
    def __init__(self, entity_json):
        self.sub_entities_to_import = []
        super().__init__('aws_iam_user', entity_json['UserName'], entity_json)

    def _generate_hcl2_code(self, entity_json) -> str:
        user_policies_code = ""
        managed_policies_code = ""

        for inline_policy in entity_json.get('UserPolicyList', []):
            inline_policy_obj = IAMInlinePolicyTransformer(inline_policy, self._safe_name, Principal.User)
            user_policies_code += inline_policy_obj.code()
            self.sub_entities_to_import += inline_policy_obj.entities_to_import()

        for managed_policy in entity_json.get('AttachedManagedPolicies', []):
            managed_policy_obj = IAMManagedPolicyAttachmentTransformer(managed_policy, self._entity_name, Principal.User)
            managed_policies_code += managed_policy_obj.code()
            self.sub_entities_to_import += managed_policy_obj.entities_to_import()

        tags = BaseEntityTransformer.transform_tags(entity_json)
        return f"""resource "aws_iam_user" "{self._safe_name}" {{
  name          = "{entity_json['UserName']}"
  path          = "{entity_json['Path']}"
  force_destroy = true
  
  {tags}
}}
{user_policies_code}{managed_policies_code}"""

    def entities_to_import(self) -> list:
        return super().entities_to_import() + self.sub_entities_to_import



import concurrent.futures
import json
import logging
import os
import pathlib
import time

import boto3
from botocore.config import Config
from botocore.exceptions import ClientError

from airiam.models.RuntimeReport import RuntimeReport

config = Config(retries={'max_attempts': 10, 'mode': 'standard'})
IAM_DATA_FILE_NAME = "iam_data.json"
ERASE_LINE = '\x1b[2K'


def get_iam_data_file(account_id: str):
    parent = f"./aircache/{account_id}"
    pathlib.Path(parent).mkdir(parents=True, exist_ok=True)
    p = f"{parent}/{IAM_DATA_FILE_NAME}"
    return p


class RuntimeIamScanner:
    """
    This class encapsulates all Runtime IAM data capture & classification
    It's entry point is the method `evaluate_runtime_iam`
    """

    def __init__(self, logger, profile=None, refresh_cache=False):
        self.logger = logger
        self.refresh_cache = refresh_cache
        if profile:
            self._session = boto3.Session(profile_name=profile)
        else:
            self._session = boto3.Session()

    def evaluate_runtime_iam(self, list_unused: bool, command: str) -> RuntimeReport:
        """
        This method encapsulates all Runtime IAM data capture & classification
        :param list_unused:  A boolean indicating whether to list the unused AWS entities or not.
        :param command:      The command inputted by the user
        :return: An instance of the report which describes which resources need to be reconfigured (and how),
                                and which resources should be removed
        """
        account_id, identity_arn = self._get_identity_details()
        iam_data = self._get_data_from_aws(account_id, list_unused)

        if command == 'terraform':
            if ':role/' in identity_arn:
                iam_data['AccountRoles'] = list(filter(
                    lambda r: r['RoleName'] != identity_arn.split('/').pop(1), iam_data['AccountRoles']))
            print(f'Filtered {identity_arn} from the analysis')

        return RuntimeReport(account_id, identity_arn, iam_data)

    def _get_data_from_aws(self, account_id: str, list_unused: bool) -> dict:
        """
        This method encapsulates all the API calls made to the AWS IAM service to gather data for later analysis
        :return: The IAM data that was pulled from the account, as was also saved locally for quicker re-runs
        """

        data_account_id = RuntimeIamScanner._get_account_id_from_existing_data(
            account_id
        )

        iam_file_name = get_iam_data_file(account_id)
        logging.debug("IAM FILE NAME", iam_file_name)
        logging.debug("Data account  id", data_account_id)

        if not self.refresh_cache and data_account_id == account_id:
            print("Reusing local data")
        else:
            print(f"Getting all IAM configurations for account {account_id}")
            iam = self._session.client('iam', config=config)
            iam.generate_credential_report()
            account_users, account_roles, account_groups, account_policies = RuntimeIamScanner.get_account_iam_configuration(
                iam)
            print("Getting IAM credential report")
            csv_credential_report = iam.get_credential_report()[
                'Content'].decode('utf-8')
            credential_report = RuntimeIamScanner.convert_csv_to_json(
                csv_credential_report)

            if list_unused:
                account_principals = account_users + account_roles
                entity_arn_list = list(
                    map(lambda e: e['Arn'], account_principals))
                last_accessed_map = self._generate_last_access(
                    iam, entity_arn_list)

                for arn, last_accessed_list in last_accessed_map.items():
                    entity = next(
                        entity for entity in account_principals if entity['Arn'] == arn)
                    entity['LastAccessed'] = last_accessed_list

            print("Collecting password configurations for all IAM users in the account")
            for user in account_users:
                try:
                    iam.get_login_profile(UserName=user['UserName'])
                    user['LoginProfileExists'] = True
                except ClientError as exception:
                    if exception.response['Error']['Code'] == 'NoSuchEntity':
                        user['LoginProfileExists'] = False
                    else:
                        raise exception

            print("Completed data collection, writing to local file...")
            iam_data = {
                'CredentialReport': credential_report,
                'AccountUsers': account_users,
                'AccountRoles': account_roles,
                'AccountGroups': account_groups,
                'AccountPolicies': account_policies
            }
            with open(get_iam_data_file(account_id=account_id), "w") as iam_file:
                json.dump(iam_data, iam_file, indent=4,
                          sort_keys=True, default=str)
        with open(get_iam_data_file(account_id=account_id)) as iam_data_file:
            iam_data = json.load(iam_data_file)

        return iam_data

    def _get_aws_iam_client(self):
        """
        Create an AWS IAM client with the profile that was supplies or default credentials if none was supplied
        :return: AWS IAM client
        """

    @staticmethod
    def get_account_iam_configuration(iam):
        marker = None
        paginator = iam.get_paginator('get_account_authorization_details')
        account_users = []
        account_roles = []
        account_policies = []
        account_groups = []
        response_iterator = paginator.paginate(
            Filter=['User', 'Role', 'Group',
                    'LocalManagedPolicy', 'AWSManagedPolicy'],
            PaginationConfig={
                'PageSize': 100,
                'StartingToken': marker
            }
        )
        for page in response_iterator:
            account_users.extend(page['UserDetailList'])
            account_roles.extend(page['RoleDetailList'])
            account_groups.extend(page['GroupDetailList'])
            account_policies.extend(page['Policies'])

        for policy in account_policies:
            policy['Description'] = iam.get_policy(PolicyArn=policy['Arn'])[
                'Policy'].get('Description', '')

        marker = None
        list_roles_result = []
        paginator = iam.get_paginator('list_roles')
        response_iterator = paginator.paginate(
            PaginationConfig={
                'PageSize': 100,
                'StartingToken': marker
            }
        )
        for page in response_iterator:
            list_roles_result.extend(page['Roles'])

        for role in account_roles:
            role['Description'] = next(role_obj.get(
                'Description', '') for role_obj in list_roles_result if role_obj['RoleName'] == role['RoleName'])

        account_policies = list(
            filter(lambda p: p['Arn'].split(':')[4] != '', account_policies))
        account_roles = list(filter(lambda r: r['Arn'].split(
            '/')[1] != 'aws-service-role', account_roles))
        return account_users, account_roles, account_groups, account_policies

    def _generate_last_access(self, iam, arn_list: list):
        results = {}
        futures = []
        print(ERASE_LINE + f"\rGenerating usage reports for {len(arn_list)} principals")

        with concurrent.futures.ThreadPoolExecutor(max_workers=os.getenv("MAX_WORKERS", 5)) as executor:
            for arn in arn_list:
                futures.append(executor.submit(
                    RuntimeIamScanner._generate_last_access_for_entity,
                    arn,
                    iam,
                    results
                ))
        concurrent.futures.wait(futures)
        print(f"\rReceived reports for {len(arn_list)} principals")
        return results

    @staticmethod
    def _generate_last_access_for_entity(arn: str, iam: boto3.client, results: dict):
        try:
            print(f"Generating report for {arn}")
            job_id = iam.generate_service_last_accessed_details(Arn=arn)['JobId']
            result = {'JobStatus': 'IN_PROGRESS'}
            for i in range(3):
                result = iam.get_service_last_accessed_details(JobId=job_id)
                if result['JobStatus'] != 'COMPLETED':
                    time.sleep(3)
                else:
                    break
            results[arn] = RuntimeIamScanner.simplify_service_access_result(result['ServicesLastAccessed'])
        except Exception as err:
            logging.error(f"Caught an error while getting the service last accessed details for {arn}, {str(err)}")
            del results[arn]
            raise err

    @staticmethod
    def convert_csv_to_json(csv_report: str):
        """
        Convert a CSV string to a json file by parsing the first row as keys, and the rows as values
        :param csv_report: a csv string, delimited with "," and rows split with "\n"
        :return: The csv as a json array
        """
        json_report = []
        rows = csv_report.split('\n')
        headers = rows[0].split(',')
        for row in rows[1:]:
            values = row.split(',')
            entity = {}
            for i in range(len(values)):
                if values[i] != 'N/A':
                    entity[headers[i]] = values[i]
            json_report.append(entity)
        return json_report

    @staticmethod
    def simplify_service_access_result(service_access_list: list):
        """
        Simplifies AWS's service_last_accessed object by keeping only the relevant fields of the services that were in use
        :param service_access_list: List of service_last_accessed objects as received from the API
        :return: A list of objects of the format {"ServiceNamespace": ..., "LastAccessed": ...}, and only those that were in use
        """
        return list(map(lambda last_access: {"ServiceNamespace": last_access["ServiceNamespace"], "LastAccessed": last_access["LastAuthenticated"]},
                        filter(lambda last_access: last_access['TotalAuthenticatedEntities'] > 0, service_access_list)))

    def _get_identity_details(self) -> (str, str):
        caller_identity_resp = self._session.client(
            'sts').get_caller_identity()
        return caller_identity_resp['Account'], caller_identity_resp['Arn']

    @staticmethod
    def _get_account_id_from_existing_data(account_id=None):
        # noinspection PyBroadException
        try:
            with open(get_iam_data_file(account_id=account_id)) as iam_data_file:
                iam_data = json.load(iam_data_file)
            return iam_data['AccountUsers'][0]['Arn'].split(":")[4]
        except Exception:
            return ""



import json

from airiam.terraform.entity_terraformers.BaseEntityTransformer import BaseEntityTransformer


class IAMPolicyDocumentTransformer(BaseEntityTransformer):
    def __init__(self, entity_json: dict, policy_name, principal_name=None):
        policy_document_name = f"{policy_name}_document"
        if principal_name:
            policy_document_name = f"{principal_name}_{policy_document_name}"
        super().__init__('data.aws_iam_policy_document', policy_document_name, entity_json)

    def _generate_hcl2_code(self, entity_json) -> str:
        statements = IAMPolicyDocumentTransformer.force_list(entity_json['Statement'])
        if 'Principal' in statements[0]:
            statements = self.transform_assume_policy_statements(statements)
        else:
            statements = self.transform_execution_policy(statements)
        code = f"""data "aws_iam_policy_document" "{self._safe_name}" {{
  version = "{entity_json.get('Version', '2012-10-17')}"
{statements}}}"""
        return code

    @staticmethod
    def transform_execution_policy(statements):
        statement_block = ""
        for statement in statements:
            sid_string = ""
            if statement.get('Sid', '') != '':
                sid_string = f"sid    = \"{statement['Sid']}\"\n    "

            actions = IAMPolicyDocumentTransformer.force_list(statement.get('Action'))
            if 'Action' in statement:
                action_str = f"actions = {json.dumps(actions)}"
            else:
                actions = IAMPolicyDocumentTransformer.force_list(statement.get('NotAction'))
                action_str = f"not_actions = {json.dumps(actions)}"
            condition_block = IAMPolicyDocumentTransformer.transform_conditions(statement)
            resources_list_str = json.dumps(IAMPolicyDocumentTransformer.force_list(statement.get('Resource'))).replace('${', '$\\u0024{')
            statement_block += f"""  statement {{
    {sid_string}effect    = "{statement['Effect']}"
    {action_str}
    resources = {resources_list_str}
    {condition_block}
  }}
"""

        return statement_block

    @staticmethod
    def transform_assume_policy_statements(statements):
        statement_block = ""
        for statement in statements:
            sid_string = ""
            if statement.get('Sid', '') != '':
                sid_string = f"sid    = \"{statement['Sid']}\"\n    "
            condition_block = IAMPolicyDocumentTransformer.transform_conditions(statement)

            statement_block += f"""  statement {{
    {sid_string}effect  = "{statement['Effect']}"
    actions = {json.dumps(IAMPolicyDocumentTransformer.force_list(statement['Action']))}
    principals {{
      type        = "{list(statement['Principal'].keys())[0]}"
      identifiers = {json.dumps(IAMPolicyDocumentTransformer.force_list(statement['Principal'][list(statement['Principal'].keys())[0]]))}
    }}
  {condition_block}}}
"""

        return statement_block

    @staticmethod
    def transform_conditions(statement):
        condition_block = ""
        if 'Condition' in statement:
            for test, items in statement['Condition'].items():
                for variable, values in items.items():
                    values_str = json.dumps(IAMPolicyDocumentTransformer.force_list(values)).replace('${', '$\\u0024{')
                    condition_block += f"""
    condition {{
      test     = "{test}"
      variable = "{variable}"
      values   = {values_str}
    }}
  """
        return condition_block

    @staticmethod
    def force_list(x):
        if isinstance(x, list):
            return x
        return [x]

    def entities_to_import(self) -> list:
        return []



import copy
import datetime as dt
import logging
from datetime import timezone

from airiam.find_unused.PolicyAnalyzer import PolicyAnalyzer
from airiam.find_unused.RuntimeIamScanner import RuntimeIamScanner


def filter_attachments_of_unused_entities(unused_policy_attachments, unused_users, unused_roles,
                                          redundant_groups) -> list:
    unused_role_names = list(map(lambda role_obj: role_obj['RoleName'], unused_roles))
    unused_user_names = list(map(lambda user_obj: user_obj['UserName'], unused_users))
    redundant_group_names = list(map(lambda group_obj: group_obj['GroupName'], redundant_groups))
    unused_policy_attachments_of_in_use_principals = []
    for policy_attachment_obj in unused_policy_attachments:
        if 'Role' in policy_attachment_obj:
            if policy_attachment_obj['Role'] not in unused_role_names:
                unused_policy_attachments_of_in_use_principals.append(policy_attachment_obj)
        elif 'User' in policy_attachment_obj:
            if policy_attachment_obj['User'] not in unused_user_names:
                unused_policy_attachments_of_in_use_principals.append(policy_attachment_obj)
        elif 'Group' in policy_attachment_obj:
            if policy_attachment_obj['Group'] not in redundant_group_names:
                unused_policy_attachments_of_in_use_principals.append(policy_attachment_obj)

    return unused_policy_attachments_of_in_use_principals


def filter_credentials_of_unused_users(unused_active_access_keys, unused_console_login_profiles, unused_users) -> (
list, list):
    result_access_keys = []
    result_console_logins = []
    for access_key in unused_active_access_keys:
        is_belong_to_unused_user = len(
            list(filter(lambda user: user['UserName'] == access_key['User'], unused_users))) > 0
        if not is_belong_to_unused_user:
            result_access_keys.append(access_key)
    for console_login in unused_console_login_profiles:
        is_belong_to_unused_user = len(
            list(filter(lambda user: user['UserName'] == console_login['User'], unused_users))) > 0
        if not is_belong_to_unused_user:
            result_console_logins.append(console_login)
    return result_access_keys, result_console_logins


def find_unused(logger, profile, refresh_cache, unused_threshold, command):
    iam_report = RuntimeIamScanner(logger, profile, refresh_cache).evaluate_runtime_iam(True, command)
    raw_iam_data = iam_report.get_raw_data()
    credential_report = raw_iam_data['CredentialReport']
    account_users = raw_iam_data['AccountUsers']
    account_roles = raw_iam_data['AccountRoles']
    account_policies = raw_iam_data['AccountPolicies']
    account_groups = raw_iam_data['AccountGroups']
    unused_users, used_users = find_unused_users(account_users, credential_report, unused_threshold)
    unused_active_access_keys, unused_console_login_profiles = find_unused_active_credentials(account_users,
                                                                                              credential_report,
                                                                                              unused_threshold)
    unattached_policies = find_unattached_policies(account_policies)
    redundant_groups = find_redundant_groups(account_groups, account_users)
    unused_roles, used_roles = find_unused_roles(account_roles, unused_threshold)
    unused_policy_attachments = find_unused_policy_attachments(account_users, account_roles, account_policies,
                                                               account_groups, unused_threshold)

    unused_access_keys, unused_console_access = filter_credentials_of_unused_users(unused_active_access_keys,
                                                                                   unused_console_login_profiles,
                                                                                   unused_users)
    unused_policy_attachments = filter_attachments_of_unused_entities(unused_policy_attachments, unused_users,
                                                                      unused_roles, redundant_groups)

    iam_report.set_unused(unused_users, unused_roles, unused_access_keys, unused_console_access, unattached_policies,
                          redundant_groups, unused_policy_attachments)
    return iam_report


def find_unused_users(users, credential_report, unused_threshold) -> (list, list):
    unused_users = []
    used_users = []
    for user in users:
        credentials = next((creds for creds in credential_report if creds['user'] == user['UserName']), {})

        findMinimumUsed = [
            days_from_today(credentials.get('access_key_1_last_used_date', 'N/A')),
            days_from_today(credentials.get('access_key_2_last_used_date', 'N/A')),
            days_from_today(credentials.get('password_last_used', 'N/A'))
        ]
        try:
            # Needs to be >= here not just >, as we need to catch "0" for accessed "today", which will fail both if
            # and elif below and put the user into "used_users" so any unused keys can be caught later.
            last_used_in_days = min(i for i in findMinimumUsed if i >= 0)
        except:
            last_used_in_days = min(i for i in findMinimumUsed if i < 0)

        if last_used_in_days >= unused_threshold:
            user['LastUsed'] = last_used_in_days
            unused_users.append(user)
        elif last_used_in_days < 0:
            user['LastUsed'] = last_used_in_days
            unused_users.append(user)
        else:
            user['LastUsed'] = last_used_in_days
            used_users.append(user)
    return unused_users, used_users


def find_unused_active_credentials(users, credential_report, unused_threshold) -> (list, list):
    unused_access_keys = []
    unused_console_login_profiles = []
    for user in users:
        try:
            credentials = next(creds for creds in credential_report if creds['user'] == user['UserName'])
        except StopIteration:
            logging.warning(f'Failed to find credentials for user {user["UserName"]}, skipping this user')
            continue
        access_key_1_unused_days = days_from_today(credentials.get('access_key_1_last_used_date', 'N/A'))
        if ((credentials['access_key_1_active'] == 'true') and (
                (access_key_1_unused_days < 0) or (access_key_1_unused_days >= unused_threshold))):
            unused_access_keys.append(
                {'User': user['UserName'], 'AccessKey': '1', 'DaysSinceLastUse': access_key_1_unused_days})

        access_key_2_unused_days = days_from_today(credentials.get('access_key_2_last_used_date', 'N/A'))
        if ((credentials['access_key_2_active'] == 'true') and (
                (access_key_2_unused_days < 0) or (access_key_2_unused_days >= unused_threshold))):
            unused_access_keys.append(
                {'User': user['UserName'], 'AccessKey': '2', 'DaysSinceLastUse': access_key_2_unused_days})

        days_since_password_last_used = days_from_today(credentials.get('password_last_used', 'N/A'))
        if ((credentials['password_enabled'] == 'true') and (
                (days_since_password_last_used < 0) or (days_since_password_last_used >= unused_threshold))):
            unused_console_login_profiles.append(
                {'User': user['UserName'], 'MFAEnabled': credentials['mfa_active'] == 'true',
                 'DaysSinceLastUse': days_since_password_last_used})
    return unused_access_keys, unused_console_login_profiles


def find_unused_roles(roles, unused_threshold) -> (list, list):
    unused_roles = []
    used_roles = []
    for role in roles:
        if role.get('LastAccessed') is None:
            used_roles.append(role)
        elif len(role['LastAccessed']) == 0:
            role['LastUsed'] = -1
            unused_roles.append(role)
        else:
            last_used = max(map(lambda last_access: last_access['LastAccessed'], role.get('LastAccessed', [])))
            role['LastUsed'] = days_from_today(last_used)
            if (role['LastUsed'] < 0) or (role['LastUsed'] >= unused_threshold):
                unused_roles.append(role)
            else:
                used_roles.append(role)
    return unused_roles, used_roles


def find_unattached_policies(policies) -> list:
    return list(filter(lambda policy: policy['AttachmentCount'] == 0 and policy['PermissionsBoundaryUsageCount'] == 0,
                       policies))


def find_redundant_groups(groups, users) -> list:
    groups_with_no_active_members = _find_groups_with_no_members(groups, users)
    groups_with_no_privilege = list(
        filter(lambda g: len(g['AttachedManagedPolicies'] + g['GroupPolicyList']) == 0, groups))
    return list({v['GroupName']: v for v in groups_with_no_active_members + groups_with_no_privilege}.values())


def _find_groups_with_no_members(group_list: list, user_list: list):
    """
    Identify groups with no members by going through the
    :param group_list: List of the groups in the account
    :param user_list:  List of the active IAM users in the account
    :return: List of groups which have no active IAM users as members
    """
    empty_groups = copy.deepcopy(group_list)
    for user in user_list:
        for group in user['GroupList']:
            try:
                group_obj = next(g for g in empty_groups if g['GroupName'] == group)
                empty_groups.remove(group_obj)
            except StopIteration:
                pass

    return empty_groups


def find_unused_policy_attachments(users: list, roles: dict, account_policies: list, account_groups: list,
                                   unused_threshold) -> list:
    unused_policy_attachments = []
    for role in roles:
        unused_policy_attachments += get_unused_role_policy_attachments(account_policies, role)

    used_group_policy_attachments = []
    potential_unused_group_policy_attachments = []
    for user in users:
        services_in_use = list(map(lambda last_access: last_access['ServiceNamespace'],
                                   filter(lambda last_access: days_from_today(
                                       last_access['LastAccessed']) > 0 < unused_threshold, user['LastAccessed'])))
        user_attached_managed_policies = copy.deepcopy(user['AttachedManagedPolicies'])
        for group_name in user['GroupList']:
            group_managed_policies = next(
                g['AttachedManagedPolicies'] for g in account_groups if g['GroupName'] == group_name)
            user_attached_managed_policies.extend(
                list(map(lambda group_policy: {**group_policy, 'Group': group_name}, group_managed_policies)))

        for policy_attachment_obj in user_attached_managed_policies:
            policy_obj = next(p for p in account_policies if policy_attachment_obj['PolicyArn'] == p['Arn'])
            policy_document = \
            next(version for version in policy_obj['PolicyVersionList'] if version['IsDefaultVersion'])['Document']
            policy_is_unused = PolicyAnalyzer.is_policy_unused(policy_document, services_in_use)
            if policy_attachment_obj.get('Group'):
                attachment_id = f'{policy_attachment_obj["PolicyName"]}/{policy_attachment_obj["Group"]}'
                if policy_is_unused:
                    potential_unused_group_policy_attachments.append({**policy_attachment_obj, 'id': attachment_id})
                else:
                    used_group_policy_attachments.append({**policy_attachment_obj, 'id': attachment_id})
            elif policy_is_unused:
                unused_policy_attachments.append({**policy_attachment_obj, 'User': user['UserName']})
    used_group_policy_attachments = {v['id']: v for v in used_group_policy_attachments}

    for policy_attachment_obj in potential_unused_group_policy_attachments:
        attachment_id = f'{policy_attachment_obj["PolicyName"]}/{policy_attachment_obj["Group"]}'
        if attachment_id not in used_group_policy_attachments:
            unused_policy_attachments.append(policy_attachment_obj)
            used_group_policy_attachments[attachment_id] = "Already added to 'unused_group_policy_attachments'"
    return unused_policy_attachments


def get_unused_role_policy_attachments(account_policies, principal):
    unused_policy_attachments = []
    if principal.get('LastAccessed') is None:
        return unused_policy_attachments
    services_last_accessed = list(map(lambda access_obj: access_obj['ServiceNamespace'], principal.get('LastAccessed')))
    for managed_policy in principal['AttachedManagedPolicies']:
        policy_obj = next(pol for pol in account_policies if pol['Arn'] == managed_policy['PolicyArn'])
        policy_document = next(version for version in policy_obj['PolicyVersionList'] if version['IsDefaultVersion'])[
            'Document']
        if PolicyAnalyzer.is_policy_unused(policy_document, services_last_accessed):
            unused_policy_attachments.append({"Role": principal['RoleName'], "PolicyArn": managed_policy['PolicyArn']})
    for inline_policy in principal.get('RolePolicyList', []):
        if PolicyAnalyzer.is_policy_unused(inline_policy['PolicyDocument'], services_last_accessed):
            unused_policy_attachments.append({"Role": principal['RoleName'], "PolicyArn": inline_policy['PolicyName']})

    return unused_policy_attachments


def days_from_today(str_date_from_today):
    if str_date_from_today in ['no_information', 'N/A']:
        return -1
    date = dt.datetime.fromisoformat(str_date_from_today)
    delta = dt.datetime.now().astimezone(timezone.utc) - date

    return delta.days



import json

from airiam.terraform.entity_terraformers.BaseEntityTransformer import BaseEntityTransformer


class IAMUserGroupMembershipTransformer(BaseEntityTransformer):
    def __init__(self, entity_json: dict, user_identifier: str):
        self._user_identifier = user_identifier
        self._user_name = entity_json['UserName']
        self._groups = entity_json['Groups']
        super().__init__('aws_iam_user_group_membership', f"{entity_json['UserName']}_group_attachment", entity_json)

    def _generate_hcl2_code(self, entity_json) -> str:
        return f"""resource {self._entity_type} "{self._safe_name}" {{
  user = {self._user_identifier}.name

  groups = {json.dumps(entity_json['Groups'])}
}}
"""

    def entities_to_import(self) -> list:
        return [{"identifier": self.identifier(), "entity": f"{self._user_name}/{'/'.join(self._groups)}"}]



import copy

from airiam.find_unused.find_unused import days_from_today
from airiam.models.RuntimeReport import RuntimeReport
from airiam.find_unused.PolicyAnalyzer import PolicyAnalyzer

ADMIN_POLICY_ARN = 'arn:aws:iam::aws:policy/AdministratorAccess'
READ_ONLY_ARN = 'arn:aws:iam::aws:policy/ReadOnlyAccess'


def recommend_groups(logger, runtime_iam_report: RuntimeReport, last_used_threshold=90, organizer=None):
    account_id = runtime_iam_report.get_raw_data()['AccountRoles'][0]['Arn'].split(":")[4]
    logger.info("Analyzing data for account {}".format(account_id))

    if not organizer:
        logger.info('Using the default UserOrganizer')
        organizer = UserOrganizer(logger, last_used_threshold)

    runtime_iam_report.set_reorg(organizer.get_user_clusters(runtime_iam_report))

    return runtime_iam_report


class UserOrganizer:
    def __init__(self, logger, unused_threshold=90):
        super().__init__()
        self.logger = logger
        self.unused_threshold = unused_threshold

    def get_user_clusters(self, runtime_report: RuntimeReport) -> dict:
        """
        Returns the user clusters based on the raw data in the runtime report
        :param runtime_report: an instance of RuntimeReport which has the raw_iam_data already set
        :return: {'Admins': {'Users': [], 'Policies': []}, 'Powerusers': {'Users': [], 'Policies': []}, 'ReadOnly': {'Users': [], 'Policies': []}}
        """
        iam_data = runtime_report.get_raw_data()
        human_users, service_users = self._separate_user_types(iam_data['AccountUsers'])
        simple_user_clusters = self._create_simple_user_clusters(human_users, iam_data['AccountGroups'], iam_data['AccountPolicies'])
        return simple_user_clusters

    def _create_simple_user_clusters(self, users, account_groups, account_policies):
        clusters = {
            'Admins': {'Policies': [ADMIN_POLICY_ARN], 'Users': []},
            'ReadOnly': {'Users': [], 'Policies': [READ_ONLY_ARN]},
            'Powerusers': {'Users': [], 'Policies': []}
        }

        policies_in_use = {}
        for user in users:
            user_attached_managed_policies = copy.deepcopy(user['AttachedManagedPolicies'])
            for group_name in user['GroupList']:
                group_managed_policies = next(g['AttachedManagedPolicies'] for g in account_groups if g['GroupName'] == group_name)
                user_attached_managed_policies.extend(group_managed_policies)
            user_attached_managed_policies = list(set(map(lambda p: p['PolicyArn'], user_attached_managed_policies)))
            user_attached_managed_policies.sort()
            if ADMIN_POLICY_ARN in user_attached_managed_policies:
                clusters['Admins']['Users'].append(user['UserName'])
            else:
                services_in_use = list(
                    map(
                        lambda last_access: last_access['ServiceNamespace'],
                        filter(
                            lambda last_access: days_from_today(last_access['LastAccessed']) < self.unused_threshold,
                            user['LastAccessed']
                        )
                    )
                )

                user_attached_managed_policies_in_use = []
                for policy_arn in user_attached_managed_policies:
                    policy_obj = next(p for p in account_policies if policy_arn == p['Arn'])
                    policy_document = next(version for version in policy_obj['PolicyVersionList'] if version['IsDefaultVersion'])['Document']
                    policy_in_use = not PolicyAnalyzer.is_policy_unused(policy_document, services_in_use)
                    if policy_in_use:
                        user_attached_managed_policies_in_use.append(policy_arn)

                user_needs_write_access = False
                for pol in user_attached_managed_policies_in_use:
                    if pol not in policies_in_use:
                        policies_in_use[pol] = 0
                    policies_in_use[pol] += 1
                    policy_obj = next(p for p in account_policies if p['Arn'] == pol)
                    policy_document = next(version for version in policy_obj['PolicyVersionList'] if version['IsDefaultVersion'])['Document']
                    if PolicyAnalyzer.policy_is_write_access(policy_document):
                        user_needs_write_access = True
                        break

                if user_needs_write_access:
                    clusters['Powerusers']['Users'].append(user['UserName'])
                else:
                    clusters['ReadOnly']['Users'].append(user['UserName'])
        policies_sorted = list({k: v for k, v in sorted(policies_in_use.items(), key=lambda item: -item[1])}.keys())

        clusters['Powerusers']['Policies'] = policies_sorted
        return clusters

    def _separate_user_types(self, account_users) -> (list, list):
        human_users = []
        service_users = []
        for user in account_users:
            if user['LastUsed'] < self.unused_threshold:
                if len(user['AttachedManagedPolicies']) == 0 and len(user['GroupList']) == 0:
                    service_users.append(user)
                else:
                    human_users.append(user)

        return human_users, service_users



from airiam.terraform.entity_terraformers.BaseEntityTransformer import BaseEntityTransformer, Principal
from airiam.terraform.entity_terraformers.IAMInlinePolicyTransformer import IAMInlinePolicyTransformer
from airiam.terraform.entity_terraformers.IAMManagedPolicyAttachmentTransformer import IAMManagedPolicyAttachmentTransformer


class IAMGroupTransformer(BaseEntityTransformer):
    def __init__(self, entity_json: dict):
        self._sub_entities_to_import = []
        super().__init__('aws_iam_group', BaseEntityTransformer.safe_name_converter(entity_json['GroupName']), entity_json)

    def _generate_hcl2_code(self, entity_json) -> str:
        group_policies = ""
        managed_policies = ""

        for inline_policy in entity_json.get('UserPolicyList', []):
            transformer = IAMInlinePolicyTransformer(inline_policy, self._safe_name, Principal.Group)
            group_policies += transformer.code()
            self._sub_entities_to_import += transformer.entities_to_import()

        for managed_policy in entity_json.get('AttachedManagedPolicies', []):
            transformer = IAMManagedPolicyAttachmentTransformer(managed_policy, self._safe_name, Principal.Group)
            managed_policies += transformer.code()
            self._sub_entities_to_import += transformer.entities_to_import()

        tags = BaseEntityTransformer.transform_tags(entity_json)
        return f"""resource "aws_iam_group" "{self._safe_name}" {{
  name = "{entity_json['GroupName']}"
  path = "{entity_json['Path']}"
}}

{group_policies}
{managed_policies}
"""

    def entities_to_import(self) -> list:
        return super().entities_to_import() + self._sub_entities_to_import



from airiam.terraform.entity_terraformers.BaseEntityTransformer import BaseEntityTransformer, Principal
from airiam.terraform.entity_terraformers.IAMManagedPolicyAttachmentTransformer import IAMManagedPolicyAttachmentTransformer
from airiam.terraform.entity_terraformers.IAMInlinePolicyTransformer import IAMInlinePolicyTransformer
from airiam.terraform.entity_terraformers.IAMPolicyDocumentTransformer import IAMPolicyDocumentTransformer
from airiam.terraform.entity_terraformers.InstancProfileTransformer import InstanceProfileTransformer


class IAMRoleTransformer(BaseEntityTransformer):
    def __init__(self, entity_json: dict):
        self._sub_entities_to_import = []
        super().__init__('aws_iam_role', BaseEntityTransformer.safe_name_converter(entity_json['RoleName']), entity_json)

    def _generate_hcl2_code(self, entity_json) -> str:
        assume_policy_document = IAMPolicyDocumentTransformer(entity_json['AssumeRolePolicyDocument'], f"{self._safe_name}_assume_role_policy")
        role_policies = ""
        for role_policy in entity_json['RolePolicyList']:
            transformer = IAMInlinePolicyTransformer(role_policy, self._safe_name, Principal.Role)
            role_policies += transformer.code()
            self._sub_entities_to_import += transformer.entities_to_import()
        for policy_attachment in entity_json['AttachedManagedPolicies']:
            transformer = IAMManagedPolicyAttachmentTransformer(policy_attachment, self._safe_name, Principal.Role)
            role_policies += transformer.code()
            self._sub_entities_to_import += transformer.entities_to_import()
        instance_profiles = ""
        for instance_profile in entity_json['InstanceProfileList']:
            transformer = InstanceProfileTransformer(instance_profile, self.identifier())
            instance_profiles += transformer.code()
            self._sub_entities_to_import += transformer.entities_to_import()

        tags = self.transform_tags(entity_json)
        permissions_boundary = ''
        if 'PermissionsBoundary' in entity_json:
            permissions_boundary = f'permissions_boundary = "{entity_json["PermissionsBoundary"]["PermissionsBoundaryArn"]}"'

        return f"""resource "aws_iam_role" "{self._safe_name}" {{
  name                  = "{entity_json['RoleName']}"
  path                  = "{entity_json['Path']}"
  description           = \"{entity_json['Description']}\"
  {permissions_boundary}
  assume_role_policy = {assume_policy_document.identifier()}.json

  {tags}
}}

{assume_policy_document.code()}

{role_policies}
{instance_profiles}"""

    def entities_to_import(self) -> list:
        return super().entities_to_import() + self._sub_entities_to_import



from enum import Enum


class Principal(Enum):
    User = 'user'
    Role = 'role'
    Group = 'group'


class BaseEntityTransformer:
    def __init__(self, entity_type: str, entity_name: str, entity_json: dict):
        self._entity_type = entity_type
        self._entity_name = entity_name
        self._safe_name = BaseEntityTransformer.safe_name_converter(entity_name)
        self._code = self._generate_hcl2_code(entity_json)

    def _generate_hcl2_code(self, entity_json) -> str:
        raise NotImplementedError()

    def entities_to_import(self) -> list:
        return [{"identifier": self.identifier(), "entity": self._entity_name}]

    def code(self) -> str:
        return self._code

    def identifier(self) -> str:
        return f"{self._entity_type}.{self._safe_name}"

    @staticmethod
    def safe_name_converter(name_str: str) -> str:
        without_special_characters = ''.join(e for e in name_str if e.isalnum() or e == '_' or e == '-')
        if without_special_characters[0].isdigit():
            return f"_{without_special_characters}"
        return without_special_characters

    @staticmethod
    def transform_tags(entity_json: dict):
        tags = entity_json.get('Tags', [])
        tags.extend([
            {"Key": "Managed by", "Value": "AirIAM by Bridgecrew"},
            {"Key": "Managed through", "Value": "Terraform"}
        ])
        tag_str = "\n".join(map(lambda tag: f"    \"{tag['Key']}\" = \"{tag['Value']}\"", tags))
        return f"""tags = {{
{tag_str}
  }}"""
